#This code is from openai baseline
#https://github.com/openai/baselines/tree/master/baselines/common/vec_env

import gym
import numpy as np
from collections import OrderedDict

class VecEnv(object):
    """
    An abstract asynchronous, vectorized environment.
    """
    def __init__(self, num_envs, observation_space, action_space):
        self.num_envs = num_envs
        self.observation_space = observation_space
        self.action_space = action_space

    def reset(self):
        """
        Reset all the environments and return an array of
        observations, or a tuple of observation arrays.
        If step_async is still doing work, that work will
        be cancelled and step_wait() should not be called
        until step_async() is invoked again.
        """
        pass

    def step_async(self, actions):
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.
        You should not call this if a step_async run is
        already pending.
        """
        pass

    def step_wait(self):
        """
        Wait for the step taken with step_async().
        Returns (obs, rews, dones, infos):
         - obs: an array of observations, or a tuple of
                arrays of observations.
         - rews: an array of rewards
         - dones: an array of "episode done" booleans
         - infos: a sequence of info objects
        """
        pass

    def close(self):
        """
        Clean up the environments' resources.
        """
        pass

    def step(self, actions):
        self.step_async(actions)
        return self.step_wait()

class DummyVecEnv(VecEnv):
    """
    VecEnv that does runs multiple environments sequentially, that is,
    the step and reset commands are send to one environment at a time.
    Useful when debugging and when num_env == 1 (in the latter case,
    avoids communication overhead)
    """
    def __init__(self, env_fns):
        """
        Arguments:

        env_fns: iterable of callables      functions that build environments
        """
        self.envs = [fn() for fn in env_fns]
        env = self.envs[0]
        VecEnv.__init__(self, len(env_fns), env.observation_space, env.action_space)
        obs_space = env.observation_space
        self.keys, shapes, dtypes = obs_space_info(obs_space)

        self.buf_obs = { k: np.zeros((self.num_envs,) + tuple(shapes[k]), dtype=dtypes[k]) for k in self.keys }
        self.buf_dones = np.zeros((self.num_envs,), dtype=np.bool)
        self.buf_rews  = np.zeros((self.num_envs,), dtype=np.float32)
        self.buf_infos = [{} for _ in range(self.num_envs)]
        self.actions = None
        self.specs = [e.spec for e in self.envs]

    def step_async(self, actions):
        listify = True
        try:
            if len(actions) == self.num_envs:
                listify = False
        except TypeError:
            pass

        if not listify:
            self.actions = actions
        else:
            assert self.num_envs == 1, "actions {} is either not a list or has a wrong size - cannot match to {} environments".format(actions, self.num_envs)
            self.actions = [actions]

    def step_wait(self):
        for e in range(self.num_envs):
            action = self.actions[e]
            if isinstance(self.envs[e].action_space, gym.spaces.Discrete):
                action = int(action)

            obs, self.buf_rews[e], self.buf_dones[e], self.buf_infos[e] = self.envs[e].step(action)
            if self.buf_dones[e]:
                obs = self.envs[e].reset()
            self._save_obs(e, obs)
        return (self._obs_from_buf(), np.copy(self.buf_rews), np.copy(self.buf_dones),
                self.buf_infos.copy())

    def reset(self):
        for e in range(self.num_envs):
            obs = self.envs[e].reset()
            self._save_obs(e, obs)
        return self._obs_from_buf()

    def _save_obs(self, e, obs):
        for k in self.keys:
            if k is None:
                self.buf_obs[k][e] = obs
            else:
                self.buf_obs[k][e] = obs[k]

    def _obs_from_buf(self):
        return dict_to_obs(copy_obs_dict(self.buf_obs))

    def get_images(self):
        return [env.render(mode='rgb_array') for env in self.envs]

    def render(self, mode='human'):
        if self.num_envs == 1:
            return self.envs[0].render(mode=mode)
        else:
            return super().render(mode=mode)
    
def copy_obs_dict(obs):
    """
    Deep-copy an observation dict.
    """
    return {k: np.copy(v) for k, v in obs.items()}


def dict_to_obs(obs_dict):
    """
    Convert an observation dict into a raw array if the
    original observation space was not a Dict space.
    """
    if set(obs_dict.keys()) == {None}:
        return obs_dict[None]
    return obs_dict


def obs_space_info(obs_space):
    """
    Get dict-structured information about a gym.Space.

    Returns:
      A tuple (keys, shapes, dtypes):
        keys: a list of dict keys.
        shapes: a dict mapping keys to shapes.
        dtypes: a dict mapping keys to dtypes.
    """
    if isinstance(obs_space, gym.spaces.Dict):
        assert isinstance(obs_space.spaces, OrderedDict)
        subspaces = obs_space.spaces
    else:
        subspaces = {None: obs_space}
    keys = []
    shapes = {}
    dtypes = {}
    for key, box in subspaces.items():
        keys.append(key)
        shapes[key] = box.shape
        dtypes[key] = box.dtype
    return keys, shapes, dtypes


def obs_to_dict(obs):
    """
    Convert an observation into a dict.
    """
    if isinstance(obs, dict):
        return obs
    return {None: obs}



import matplotlib.pyplot as plt
import random
import numpy as np

plt.ion()

#from controller import Player
from env.controller import Player

class Terrain:
    def __init__(self):
        self.reward_locs = [[5.0, 5.0], [-5.0, 5.0], [5.0, -5.0], [-5.0, -5.0]]
        self.reward_range = 1.0
        self.reward_goal = 10.0
        self.bounds_x = [-8.0, 8.0]
        self.bounds_y = [-8.0, 8.0]
        self.player = Player(0.0, 0.0, self)

    def getreward(self):
        reward = 0.0
        for x_pos, y_pos in self.reward_locs:
            max_score_temp = -(np.sqrt((self.player.x - x_pos) ** 2 + (self.player.y - y_pos) ** 2) / 20)
            reward += max_score_temp

        for x_pos, y_pos in self.reward_locs:
            #reward -= 0.15
            if abs(self.player.x - x_pos) < self.reward_range and abs(self.player.y - y_pos) < self.reward_range:
                reward += self.reward_goal
        return reward

    def checkepisodeend(self):
        for x_pos, y_pos in self.reward_locs:
            if abs(self.player.x - x_pos) < self.reward_range and abs(self.player.y - y_pos) < self.reward_range:
                return 1
        return 0

    def plotgame(self):
        plt.clf()
        for x_pos, y_pos in self.reward_locs:
            plt.plot([x_pos,], [y_pos,], marker='o', markersize=10, color="green")
        plt.xlim([-8, 8])
        plt.ylim([-8, 8])
        plt.plot([self.player.x,], [self.player.y,], marker='x', markersize=3, color="red")
        plt.pause(0.001)

    def resetgame(self):
        self.player = Player(0.0, 0.0, self)

"""
terrain = Terrain()
while(1):
    terrain.plotgame()
    print terrain.player.action(random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0))
"""



import torch
import torch.autograd as autograd
import torch.nn as nn


class Policy(nn.Module):
    def __init__(self, num_inputs, num_outputs):
        super(Policy, self).__init__()
        self.affine1 = nn.Linear(num_inputs, 64)
        self.affine2 = nn.Linear(64, 64)

        self.action_mean = nn.Linear(64, num_outputs)
        self.action_mean.weight.data.mul_(0.1)
        self.action_mean.bias.data.mul_(0.0)

        self.action_log_std = nn.Parameter(torch.zeros(1, num_outputs))

        self.saved_actions = []
        self.rewards = []
        self.final_value = 0

    def forward(self, x):
        x = torch.tanh(self.affine1(x))
        x = torch.tanh(self.affine2(x))

        action_mean = self.action_mean(x)
        action_log_std = self.action_log_std.expand_as(action_mean)
        action_std = torch.exp(action_log_std)

        return action_mean, action_log_std, action_std


class Value(nn.Module):
    def __init__(self, num_inputs):
        super(Value, self).__init__()
        self.affine1 = nn.Linear(num_inputs, 64)
        self.affine2 = nn.Linear(64, 64)
        self.value_head = nn.Linear(64, 1)
        self.value_head.weight.data.mul_(0.1)
        self.value_head.bias.data.mul_(0.0)

    def forward(self, x):
        x = torch.tanh(self.affine1(x))
        x = torch.tanh(self.affine2(x))

        state_values = self.value_head(x)
        return state_values



import torch
import torch.nn.functional as F
from torch.distributions.categorical import Categorical
import gym
from tensorboardX import SummaryWriter

writer = SummaryWriter('log')

class Policy(torch.nn.Module):
    def __init__(self, n_state, n_action):
        super(Policy, self).__init__()
        self.fc1 = torch.nn.Linear(n_state, 50)
        self.fc2 = torch.nn.Linear(50, n_action)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.softmax(self.fc2(x), dim=1)
        return x

class Agent():
    def __init__(self, n_state, n_action, lr=0.01, gamma=0.99):
        self.n_state = n_state
        self.n_action = n_action
        self.policy = Policy(n_state, n_action)
        self._reset_memory()
        self.gamma = gamma
        self.op = torch.optim.Adam(self.policy.parameters(), lr=lr)
        writer.add_graph(self.policy, torch.zeros((1, n_state)))

    def get_action(self, state):
        actions = self.policy(torch.Tensor([state]))
        return Categorical(actions).sample().item()

    def push_memory(self, state, action, reward):
        self.state_memory.append(state)
        self.action_memory.append(action)
        self.reward_memory.append(reward)

    def _reset_memory(self):
        self.state_memory = []
        self.reward_memory = []
        self.action_memory = []

    def _backward_reward(self):
        reward = 0
        length = len(self.reward_memory)
        for i in range(length):
            self.reward_memory[length - 1 -i] += self.gamma * reward
            reward = self.reward_memory[length - 1 -i]

    def _learn(self):
        state = torch.Tensor(self.state_memory)  # (steps, n_state)
        action = torch.LongTensor(self.action_memory)  # (steps, )
        reward = torch.Tensor(self.reward_memory)  # (steps, )
        actions_prob = self.policy(state)  # (steps, n_action)
        action_take_prob = actions_prob.gather(1, action.unsqueeze(1)).squeeze(1) # (steps, )
        loss = torch.sum(torch.log(action_take_prob) * reward * -1)

        self.op.zero_grad()
        loss.backward()
        self.op.step()

    def learn(self):
        self._backward_reward()
        self._learn()
        self._reset_memory()

env = gym.make('CartPole-v1').unwrapped
env._max_episode_steps = 4000
N_ACTION = 2

agent = Agent(n_action=N_ACTION, n_state=env.observation_space.shape[0])
def learn_loop():
    for epoch in range(1000):
        env.reset()
        state, reward, done, info = env.step(env.action_space.sample())
        steps = 0
        while True:
            env.render()
            
            action = agent.get_action(state)
            state_, reward, done, info = env.step(action)

            agent.push_memory(state, action, reward)
            state = state_

            steps += 1
            if done:
                agent.learn()
                break
        if epoch % 5 == 0:
            print("epoch %d, steps %d" % (epoch, steps))



import math
import gym
import numpy as np
import matplotlib.pyplot as plt

import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd
import torch.nn.functional as F

from rllite.common import PrioritizedReplayBuffer
from rllite.common import make_atari, wrap_deepmind, wrap_pytorch


class NoisyLinear(nn.Module):
    def __init__(self, in_features, out_features, std_init=0.4):
        super(NoisyLinear, self).__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.std_init = std_init

        self.weight_mu = nn.Parameter(torch.FloatTensor(out_features, in_features))
        self.weight_sigma = nn.Parameter(torch.FloatTensor(out_features, in_features))
        self.register_buffer('weight_epsilon', torch.FloatTensor(out_features, in_features))

        self.bias_mu = nn.Parameter(torch.FloatTensor(out_features))
        self.bias_sigma = nn.Parameter(torch.FloatTensor(out_features))
        self.register_buffer('bias_epsilon', torch.FloatTensor(out_features))

        self.reset_parameters()
        self.reset_noise()

    def forward(self, x):
        if self.training:
            weight = self.weight_mu + self.weight_sigma.mul(self.weight_epsilon)
            bias = self.bias_mu + self.bias_sigma.mul(self.bias_epsilon)
        else:
            weight = self.weight_mu
            bias = self.bias_mu

        return F.linear(x, weight, bias)

    def reset_parameters(self):
        mu_range = 1 / math.sqrt(self.weight_mu.size(1))

        self.weight_mu.data.uniform_(-mu_range, mu_range)
        self.weight_sigma.data.fill_(self.std_init / math.sqrt(self.weight_sigma.size(1)))

        self.bias_mu.data.uniform_(-mu_range, mu_range)
        self.bias_sigma.data.fill_(self.std_init / math.sqrt(self.bias_sigma.size(0)))

    def reset_noise(self):
        epsilon_in = self._scale_noise(self.in_features)
        epsilon_out = self._scale_noise(self.out_features)

        self.weight_epsilon.copy_(epsilon_out.ger(epsilon_in))
        self.bias_epsilon.copy_(self._scale_noise(self.out_features))

    def _scale_noise(self, size):
        x = torch.randn(size)
        x = x.sign().mul(x.abs().sqrt())
        return x


class TinyNoisyDQN(nn.Module):
    def __init__(self, num_inputs, num_actions):
        super(TinyNoisyDQN, self).__init__()

        self.num_input = num_inputs
        self.num_actions = num_actions

        self.linear = nn.Linear(self.num_input, 128)
        self.noisy1 = NoisyLinear(128, 128)
        self.noisy2 = NoisyLinear(128, self.num_actions)

    def forward(self, x):
        x = F.relu(self.linear(x))
        x = F.relu(self.noisy1(x))
        x = self.noisy2(x)
        return x

    def act(self, state):
        state = torch.FloatTensor(state).unsqueeze(0)
        q_value = self.forward(state)
        action = torch.argmax(q_value, dim=1).item()
        return action

    def reset_noise(self):
        self.noisy1.reset_noise()
        self.noisy2.reset_noise()


class NoisyDQN(object):
    def __init__(self, env_id="CartPole-v0"):
        self.env_id = "CartPole-v0"
        self.env = gym.make(self.env_id)
        self.current_model = TinyNoisyDQN(self.env.observation_space.shape[0], self.env.action_space.n)
        self.target_model = TinyNoisyDQN(self.env.observation_space.shape[0], self.env.action_space.n)
        if torch.cuda.is_available():
            self.current_model = self.current_model.cuda()
            self.target_model = self.target_model.cuda()

        self.optimizer = optim.Adam(self.current_model.parameters(), lr=0.0001)
        self.replay_buffer = PrioritizedReplayBuffer(10000, alpha=0.6)

        self.update_target(self.current_model, self.target_model)
        self.losses = []

    def update_target(self, current_model, target_model):
        target_model.load_state_dict(current_model.state_dict())

    def train_step(self, frame_idx, beta_start=0.4, beta_frames=1000, gamma = 0.99, batch_size=32):
        beta = min(1.0, beta_start + frame_idx * (1.0 - beta_start) / beta_frames)
        state, action, reward, next_state, done, weights, indices = self.replay_buffer.sample(batch_size, beta)

        state = torch.FloatTensor(np.float32(state))
        next_state = torch.FloatTensor(np.float32(next_state))
        action = torch.LongTensor(action)
        reward = torch.FloatTensor(reward)
        done = torch.FloatTensor(np.float32(done))
        weights = torch.FloatTensor(weights)

        q_values = self.current_model(state)
        next_q_values = self.target_model(next_state)

        q_value = q_values.gather(1, action.unsqueeze(1)).squeeze(1)
        next_q_value = next_q_values.max(1)[0]
        expected_q_value = reward + gamma * next_q_value * (1 - done)

        loss = (q_value - expected_q_value.detach()).pow(2) * weights
        prios = loss + 1e-5
        loss = loss.mean()

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.replay_buffer.update_priorities(indices, prios.data.cpu().numpy())
        self.current_model.reset_noise()
        self.target_model.reset_noise()

        self.losses.append(loss.item())

    def learn(self, num_frames=10000, batch_size=32):
        all_rewards = []
        episode_reward = 0

        state = self.env.reset()
        for frame_idx in range(1, num_frames + 1):
            action = self.current_model.act(state)

            next_state, reward, done, _ = self.env.step(action)
            self.replay_buffer.push(state, action, reward, next_state, done)

            state = next_state
            episode_reward += reward

            if done:
                state = self.env.reset()
                all_rewards.append(episode_reward)
                episode_reward = 0

            if len(self.replay_buffer) > batch_size:
                self.train_step(frame_idx)

            # if frame_idx % 200 == 0:
            if frame_idx == num_frames:
                plt.figure(figsize=(20, 5))
                plt.subplot(121)
                plt.title('frame %s. reward: %s' % (frame_idx, np.mean(all_rewards[-10:])))
                plt.plot(all_rewards)
                plt.subplot(122)
                plt.title('loss')
                plt.plot(self.losses)
                plt.show()

            if frame_idx % 1000 == 0:
                self.update_target(self.current_model, self.target_model)

            print(frame_idx)


class TinyNoisyCnnDQN(nn.Module):
    def __init__(self, input_shape, num_actions):
        super(TinyNoisyCnnDQN, self).__init__()

        self.input_shape = input_shape
        self.num_actions = num_actions

        self.features = nn.Sequential(
            nn.Conv2d(input_shape[0], 32, kernel_size=8, stride=4),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=4, stride=2),
            nn.ReLU(),
            nn.Conv2d(64, 64, kernel_size=3, stride=1),
            nn.ReLU()
        )

        self.noisy1 = NoisyLinear(self.feature_size(), 512)
        self.noisy2 = NoisyLinear(512, self.num_actions)

    def forward(self, x):
        batch_size = x.size(0)

        x = x / 255.
        x = self.features(x)
        x = x.view(batch_size, -1)

        x = F.relu(self.noisy1(x))
        x = self.noisy2(x)
        return x

    def reset_noise(self):
        self.noisy1.reset_noise()
        self.noisy2.reset_noise()

    def feature_size(self):
        return self.features(autograd.Variable(torch.zeros(1, *self.input_shape))).view(1, -1).size(1)

    def act(self, state):
        state = torch.FloatTensor(np.float32(state)).unsqueeze(0)
        q_value = self.forward(state)
        action = torch.argmax(q_value, dim=1).item()
        return action


class NoisyCnnDQN(object):
    def __init__(self, env_id="PongNoFrameskip-v4", ):
        self.env_id = env_id
        self.env = make_atari(self.env_id)
        self.env = wrap_deepmind(self.env)
        self.env = wrap_pytorch(self.env)

        self.current_model = TinyNoisyCnnDQN(self.env.observation_space.shape, self.env.action_space.n)
        self.target_model = TinyNoisyCnnDQN(self.env.observation_space.shape, self.env.action_space.n)
        if torch.cuda.is_available():
            self.current_model = self.current_model.cuda()
            self.target_model = self.target_model.cuda()

        self.optimizer = optim.Adam(self.current_model.parameters(), lr=0.0001)
        self.replay_buffer = PrioritizedReplayBuffer(10000, alpha=0.6)
        self.losses = []

    def update_target(self, current_model, target_model):
        target_model.load_state_dict(current_model.state_dict())

    def train_step(self, frame_idx, gamma=0.99, batch_size=32, beta_start=0.4, beta_frames=100000):
        beta = min(1.0, beta_start + frame_idx * (1.0 - beta_start) / beta_frames)
        state, action, reward, next_state, done, weights, indices = self.replay_buffer.sample(batch_size, beta)

        state = torch.FloatTensor(np.float32(state))
        next_state = torch.FloatTensor(np.float32(next_state))
        action = torch.LongTensor(action)
        reward = torch.FloatTensor(reward)
        done = torch.FloatTensor(np.float32(done))
        weights = torch.FloatTensor(weights)

        q_values = self.current_model(state)
        next_q_values = self.target_model(next_state)

        q_value = q_values.gather(1, action.unsqueeze(1)).squeeze(1)
        next_q_value = next_q_values.max(1)[0]
        expected_q_value = reward + gamma * next_q_value * (1 - done)

        loss = (q_value - expected_q_value.detach()).pow(2) * weights
        prios = loss + 1e-5
        loss = loss.mean()

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.replay_buffer.update_priorities(indices, prios.data.cpu().numpy())
        self.current_model.reset_noise()
        self.target_model.reset_noise()

        self.losses.append(loss.item())

    def learn(self, num_frames=1000000, batch_size=32):
        all_rewards = []
        episode_reward = 0

        state = self.env.reset()
        for frame_idx in range(1, num_frames + 1):
            action = self.current_model.act(state)

            next_state, reward, done, _ = self.env.step(action)
            self.replay_buffer.push(state, action, reward, next_state, done)

            state = next_state
            episode_reward += reward

            if done:
                state = self.env.reset()
                all_rewards.append(episode_reward)
                episode_reward = 0

            if len(self.replay_buffer) > batch_size:
                self.train_step(frame_idx)

            if frame_idx % 10000 == 0:
                plt.figure(figsize=(20, 5))
                plt.subplot(121)
                plt.title('frame %s. reward: %s' % (frame_idx, np.mean(all_rewards[-10:])))
                plt.plot(all_rewards)
                plt.subplot(122)
                plt.title('loss')
                plt.plot(self.losses)
                plt.show()

            if frame_idx % 1000 == 0:
                self.update_target(self.current_model, self.target_model)

            print(frame_idx)


if __name__ == '__main__':
    model = NoisyDQN()
    model.learn()
    model2 = NoisyCnnDQN()
    model2.learn()



import numpy as np
from collections import deque
import gym
from gym import spaces
import cv2

cv2.ocl.setUseOpenCL(False)


class NoopResetEnv(gym.Wrapper):
    def __init__(self, env, noop_max=30):
        """Sample initial states by taking random number of no-ops on reset.
        No-op is assumed to be action 0.
        """
        gym.Wrapper.__init__(self, env)
        self.noop_max = noop_max
        self.override_num_noops = None
        self.noop_action = 0
        assert env.unwrapped.get_action_meanings()[0] == 'NOOP'

    def reset(self, **kwargs):
        """ Do no-op action for a number of steps in [1, noop_max]."""
        self.env.reset(**kwargs)
        if self.override_num_noops is not None:
            noops = self.override_num_noops
        else:
            noops = self.unwrapped.np_random.randint(1, self.noop_max + 1)  # pylint: disable=E1101
        assert noops > 0
        obs = None
        for _ in range(noops):
            obs, _, done, _ = self.env.step(self.noop_action)
            if done:
                obs = self.env.reset(**kwargs)
        return obs

    def step(self, ac):
        return self.env.step(ac)


class FireResetEnv(gym.Wrapper):
    def __init__(self, env):
        """Take action on reset for environments that are fixed until firing."""
        gym.Wrapper.__init__(self, env)
        assert env.unwrapped.get_action_meanings()[1] == 'FIRE'
        assert len(env.unwrapped.get_action_meanings()) >= 3

    def reset(self, **kwargs):
        self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(1)
        if done:
            self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(2)
        if done:
            self.env.reset(**kwargs)
        return obs

    def step(self, ac):
        return self.env.step(ac)


class EpisodicLifeEnv(gym.Wrapper):
    def __init__(self, env):
        """Make end-of-life == end-of-episode, but only reset on true game over.
        Done by DeepMind for the DQN and co. since it helps value estimation.
        """
        gym.Wrapper.__init__(self, env)
        self.lives = 0
        self.was_real_done = True

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.was_real_done = done
        # check current lives, make loss of life terminal,
        # then update lives to handle bonus lives
        lives = self.env.unwrapped.ale.lives()
        if lives < self.lives and lives > 0:
            # for Qbert sometimes we stay in lives == 0 condtion for a few frames
            # so its important to keep lives > 0, so that we only reset once
            # the environment advertises done.
            done = True
        self.lives = lives
        return obs, reward, done, info

    def reset(self, **kwargs):
        """Reset only when lives are exhausted.
        This way all states are still reachable even though lives are episodic,
        and the learner need not know about any of this behind-the-scenes.
        """
        if self.was_real_done:
            obs = self.env.reset(**kwargs)
        else:
            # no-op step to advance from terminal/lost life state
            obs, _, _, _ = self.env.step(0)
        self.lives = self.env.unwrapped.ale.lives()
        return obs


class MaxAndSkipEnv(gym.Wrapper):
    def __init__(self, env, skip=4):
        """Return only every `skip`-th frame"""
        gym.Wrapper.__init__(self, env)
        # most recent raw observations (for max pooling across time steps)
        self._obs_buffer = np.zeros((2,) + env.observation_space.shape, dtype=np.uint8)
        self._skip = skip

    def reset(self):
        return self.env.reset()

    def step(self, action):
        """Repeat action, sum reward, and max over last observations."""
        total_reward = 0.0
        done = None
        for i in range(self._skip):
            obs, reward, done, info = self.env.step(action)
            if i == self._skip - 2: self._obs_buffer[0] = obs
            if i == self._skip - 1: self._obs_buffer[1] = obs
            total_reward += reward
            if done:
                break
        # Note that the observation on the done=True frame
        # doesn't matter
        max_frame = self._obs_buffer.max(axis=0)

        return max_frame, total_reward, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)


class ClipRewardEnv(gym.RewardWrapper):
    def __init__(self, env):
        gym.RewardWrapper.__init__(self, env)

    def reward(self, reward):
        """Bin reward to {+1, 0, -1} by its sign."""
        return np.sign(reward)


class WarpFrame(gym.ObservationWrapper):
    def __init__(self, env):
        """Warp frames to 84x84 as done in the Nature paper and later work."""
        gym.ObservationWrapper.__init__(self, env)
        self.width = 84
        self.height = 84
        self.observation_space = spaces.Box(low=0, high=255,
                                            shape=(self.height, self.width, 1), dtype=np.uint8)

    def observation(self, frame):
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
        frame = cv2.resize(frame, (self.width, self.height), interpolation=cv2.INTER_AREA)
        return frame[:, :, None]


class FrameStack(gym.Wrapper):
    def __init__(self, env, k):
        """Stack k last frames.
        Returns lazy array, which is much more memory efficient.
        See Also
        --------
        baselines.common.atari_wrappers.LazyFrames
        """
        gym.Wrapper.__init__(self, env)
        self.k = k
        self.frames = deque([], maxlen=k)
        shp = env.observation_space.shape
        self.observation_space = spaces.Box(low=0, high=255, shape=(shp[0], shp[1], shp[2] * k), dtype=np.uint8)

    def reset(self):
        ob = self.env.reset()
        for _ in range(self.k):
            self.frames.append(ob)
        return self._get_ob()

    def step(self, action):
        ob, reward, done, info = self.env.step(action)
        self.frames.append(ob)
        return self._get_ob(), reward, done, info

    def _get_ob(self):
        assert len(self.frames) == self.k
        return LazyFrames(list(self.frames))


class ScaledFloatFrame(gym.ObservationWrapper):
    def __init__(self, env):
        gym.ObservationWrapper.__init__(self, env)

    def observation(self, observation):
        # careful! This undoes the memory optimization, use
        # with smaller replay buffers only.
        return np.array(observation).astype(np.float32) / 255.0


class LazyFrames(object):
    def __init__(self, frames):
        """This object ensures that common frames between the observations are only stored once.
        It exists purely to optimize memory usage which can be huge for DQN's 1M frames replay
        buffers.
        This object should only be converted to numpy array before being passed to the model.
        You'd not believe how complex the previous solution was."""
        self._frames = frames
        self._out = None

    def _force(self):
        if self._out is None:
            self._out = np.concatenate(self._frames, axis=2)
            self._frames = None
        return self._out

    def __array__(self, dtype=None):
        out = self._force()
        if dtype is not None:
            out = out.astype(dtype)
        return out

    def __len__(self):
        return len(self._force())

    def __getitem__(self, i):
        return self._force()[i]


def make_atari(env_id):
    env = gym.make(env_id)
    assert 'NoFrameskip' in env.spec.id
    env = NoopResetEnv(env, noop_max=30)
    env = MaxAndSkipEnv(env, skip=4)
    return env


def wrap_deepmind(env, episode_life=True, clip_rewards=True, frame_stack=False, scale=False):
    """Configure environment for DeepMind-style Atari.
    """
    if episode_life:
        env = EpisodicLifeEnv(env)
    if 'FIRE' in env.unwrapped.get_action_meanings():
        env = FireResetEnv(env)
    env = WarpFrame(env)
    if scale:
        env = ScaledFloatFrame(env)
    if clip_rewards:
        env = ClipRewardEnv(env)
    if frame_stack:
        env = FrameStack(env, 4)
    return env


class ImageToPyTorch(gym.ObservationWrapper):
    """
    Image shape to num_channels x weight x height
    """

    def __init__(self, env):
        super(ImageToPyTorch, self).__init__(env)
        old_shape = self.observation_space.shape
        self.observation_space = gym.spaces.Box(low=0.0, high=1.0, shape=(old_shape[-1], old_shape[0], old_shape[1]),
                                                dtype=np.uint8)

    def observation(self, observation):
        return np.swapaxes(observation, 2, 0)


def wrap_pytorch(env):
    return ImageToPyTorch(env)



class Player:
    def __init__(self, x, y, terrain):
        self.x = x
        self.y = y
        self.terrain = terrain

    def getposition(self):
        return (self.x, self.y)

    def action(self, xy_m):
        self.x += xy_m[0]
        self.y += xy_m[1]
        self.x = min(self.terrain.bounds_x[1], max(self.x, self.terrain.bounds_x[0]))
        self.y = min(self.terrain.bounds_y[1], max(self.y, self.terrain.bounds_y[0]))
        return self.terrain.getreward()



import random
import numpy as np

import torch
from torch.autograd import Variable
import torch.nn as nn
import torch.nn.functional as F

from qnet import QNet
from policynet import PolicyNet

from env.terrain import Terrain

class SoftQLearning:
    def __init__(self):
        self.dO = 2 # Observation space dimensions
        self.dA = 2 # Action space dimensions

        self.criterion = nn.MSELoss()

        self.q_net = QNet()
        self.q_optimizer = torch.optim.Adam(self.q_net.parameters())
        self.policy_net = PolicyNet()
        self.policy_optimizer = torch.optim.Adam(self.policy_net.parameters())
        self.terrain = Terrain()
 
        self.replay_buffer_maxlen = 50
        self.replay_buffer = []
        self.exploration_prob = 0.0
        self.alpha = 1.0
        self.value_alpha = 0.3

        self.action_set = []
        for j in range(32):
            self.action_set.append((np.sin((3.14*2/32)*j), np.cos((3.14*2/32)*j)))

    def forward_QNet(self, obs, action):
        inputs = Variable(torch.FloatTensor([obs + action]))
        q_pred = self.q_net(inputs)
        return q_pred

    def forward_PolicyNet(self, obs, noise):
        inputs = Variable(torch.FloatTensor([obs + noise]))
        action_pred = self.policy_net(inputs)
        return action_pred

    def collect_samples(self):
        self.replay_buffer = []
        self.terrain.resetgame()
        while(1):
            self.terrain.plotgame()
            current_state = self.terrain.player.getposition()
            """
            best_action = self.action_set[0]
            for j in range(32):
                # Sample 32 actions and use them in the next state to get maximum Q_value
                action_temp = self.action_set[j]
                print self.forward_QNet(current_state, action_temp).data.numpy()[0][0]
                if self.forward_QNet(current_state, action_temp).data.numpy()[0][0] > self.forward_QNet(current_state, best_action).data.numpy()[0][0]:
                    best_action = action_temp
            print "Exploration prob:", self.exploration_prob
            """
            best_action = tuple(self.forward_PolicyNet(current_state, (np.random.normal(0.0, 0.5), np.random.normal(0.0, 0.5))).data.numpy()[0].tolist())
            if random.uniform(0.0, 1.0) < self.exploration_prob:
                x_val = random.uniform(-1.0, 1.0)
                best_action = (x_val, random.choice([-1.0, 1.0])*np.sqrt(1.0 - x_val*x_val))
            print("Action:", best_action)
            current_reward = self.terrain.player.action(best_action)
            print("Reward:", current_reward)
            next_state = self.terrain.player.getposition()
            self.replay_buffer.append([current_state, best_action, current_reward, next_state])
            if self.terrain.checkepisodeend() or len(self.replay_buffer) > self.replay_buffer_maxlen:
                self.terrain.resetgame()
                break

    def rbf_kernel(self, input1, input2):
        return np.exp(-3.14*(np.dot(input1-input2,input1-input2)))

    def rbf_kernel_grad(self, input1, input2):
        diff = (input1-input2)
        mult_val = self.rbf_kernel(input1, input2) * -2 * 3.14 
        return [x * mult_val for x in diff]

    def train_network(self):
        for t in range(50):
            i = random.randint(0, len(self.replay_buffer)-1)
            current_state = self.replay_buffer[i][0]
            current_action = self.replay_buffer[i][1]
            current_reward = self.replay_buffer[i][2]
            next_state = self.replay_buffer[i][3]

            # Perform updates on the Q-Network
            best_q_val_next = 0
            for j in range(32):
                # Sample 32 actions and use them in the next state to get an estimate of the state value
                action_temp = self.action_set[j]
                q_value_temp = (1.0/self.value_alpha) * self.forward_QNet(next_state, action_temp).data.numpy()[0][0]
                q_value_temp = np.exp(q_value_temp) / (1.0/32)
                best_q_val_next += q_value_temp * (1.0/32)
            best_q_val_next = self.value_alpha * np.log(best_q_val_next)
            print("Best Q Val:", best_q_val_next)
            inputs_cur = Variable(torch.FloatTensor([(current_state + current_action)]), requires_grad=True)
            predicted_q = self.q_net(inputs_cur)
            expected_q = current_reward + 0.99 * best_q_val_next
            expected_q = (1-self.alpha) * predicted_q.data.numpy()[0][0] + self.alpha * expected_q
            expected_q = Variable(torch.FloatTensor([[expected_q]]))
            loss = self.criterion(predicted_q, expected_q)
            loss.backward()

            # Perform updates on the Policy-Network using SVGD
            action_predicted = self.forward_PolicyNet(current_state, (0.0, 0.0))
            final_action_gradient = [0.0, 0.0]
            for j in range(32):
                action_temp = tuple(self.forward_PolicyNet(current_state, (np.random.normal(0.0, 0.5), np.random.normal(0.0, 0.5))).data.numpy()[0].tolist())
                inputs_temp = Variable(torch.FloatTensor([current_state + action_temp]), requires_grad=True)
                predicted_q = self.q_net(inputs_temp)

                # Perform standard Q-value computation for each of the selected actions
                best_q_val_next = 0
                for k in range(32):
                    # Sample 32 actions and use them in the next state to get an estimate of the state value
                    action_temp_2 = self.action_set[k]
                    q_value_temp = (1.0/self.value_alpha) * self.forward_QNet(next_state, action_temp_2).data.numpy()[0][0]
                    q_value_temp = np.exp(q_value_temp) / (1.0/32)
                    best_q_val_next += q_value_temp * (1.0/32)
                best_q_val_next = self.value_alpha * np.log(best_q_val_next)
                expected_q = current_reward + 0.99 * best_q_val_next
                expected_q = (1-self.alpha) * predicted_q.data.numpy()[0][0] + self.alpha * expected_q
                expected_q = Variable(torch.FloatTensor([[expected_q]]))
                loss = self.criterion(predicted_q, expected_q)
                loss.backward()

                action_gradient_temp = [inputs_temp.grad.data.numpy()[0][2], inputs_temp.grad.data.numpy()[0][3]]
                kernel_val = self.rbf_kernel(list(action_temp), action_predicted.data.numpy()[0])
                kernel_grad = self.rbf_kernel_grad(list(action_temp), action_predicted.data.numpy()[0])
                final_temp_grad = ([x * kernel_val for x in action_gradient_temp] + [x * self.value_alpha for x in kernel_grad])
                final_action_gradient[0] += (1.0/32) * final_temp_grad[0]
                final_action_gradient[1] += (1.0/32) * final_temp_grad[1]
            print(final_action_gradient)
            action_predicted.backward(torch.FloatTensor([final_action_gradient]))

            # Apply the updates using the optimizers
            self.q_optimizer.zero_grad()
            self.q_optimizer.step()
            self.policy_optimizer.zero_grad()
            self.policy_optimizer.step()


softqlearning = SoftQLearning()
while(1):
    if softqlearning.exploration_prob > 0.1:
        softqlearning.exploration_prob *= 0.9
    softqlearning.collect_samples()
    softqlearning.train_network()
